package sd.computer;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import sd.constants.Constants;
import sd.data.Database;

public class KLDiv {
	private static final double EPSILON = 0.00005;
	
	public static double score(Database db, List<Integer> pids) {
		HashMap<String[], Integer> stats = new HashMap<String[], Integer>();
		for (Integer id : pids) {
			String[] key = db.getPoints().get(id).getCategoricTarget();
			if (stats.containsKey(key)) {
				int a = stats.get(key) + 1;
				stats.put(key, a);
			} else {
				stats.put(key, 1);
			}
		}
		
		if (stats.keySet().size() == 0) {
			throw new ArithmeticException("invalid frequency computation!");
		}
		
		Map<String[], Integer> pops = db.getCategoricTargetCount();
		double sum = 0;
		for (String[] key : stats.keySet()) {
			double a = stats.get(key) * 1.0 / pids.size();
			double b = pops.get(key) * 1.0 / db.getPoints().size();
			if (a == 0) {
				throw new ArithmeticException("invalid frequency computation!");
			}
			sum += a * Math.log(a / b) / Math.log(2); 
		}
		
		if (sum < 0) {
			throw new ArithmeticException("invalid score computation!");
		}
		
		//System.out.printf("score = %f\n", sum);
		
		return sum;
	}
	
	public static double score(Database db, List<Integer> a, List<Integer> b) {
		HashMap<String[], Integer> statsA = getMap(db, a);
		HashMap<String[], Integer> statsB = getMap(db, b);
		
		if (statsA.keySet().size() == 0 || statsB.keySet().size() == 0) {
			throw new ArithmeticException("invalid frequency computation!");
		}
		
		double scoreAtoB = KLScore(statsA, a.size(), statsB, b.size());
		double scoreBtoA = KLScore(statsB, b.size(), statsA, a.size());
		
		return scoreAtoB + scoreBtoA;
	}
	
	private static double KLScore(HashMap<String[], Integer> statsA, int sizeA, HashMap<String[], Integer> statsB, int sizeB) {
		double sum = 0;
		for (String[] key : statsA.keySet()) {
			double a = statsA.get(key) * 1.0 / sizeA;
			if (a == 0) {
				throw new ArithmeticException("invalid frequency computation!");
			}
			
			double b = 0;
			if (statsB.containsKey(key)) {
				b = statsB.get(key) * 1.0 / sizeB;
			}
			
			sum += (a + EPSILON) * Math.log((a + EPSILON) / (b + EPSILON)) / Math.log(2);
		}
		
		if (sum < 0) {
			throw new ArithmeticException("invalid score computation!");
		}
		
		return sum;
	}
	
	public static HashMap<String[], Integer> getMap(Database db, List<Integer> pids) {
		HashMap<String[], Integer> stats = new HashMap<String[], Integer>();
		for (Integer id : pids) {
			String[] key = db.getPoints().get(id).getCategoricTarget();
			if (stats.containsKey(key)) {
				int a = stats.get(key) + 1;
				stats.put(key, a);
			} else {
				stats.put(key, 1);
			}
		}
		
		return stats;
	}
	
	@SuppressWarnings("unchecked")
	public static double scoreIndependence(Database db, List<Integer> pids) {
		HashMap<String, Integer>[] stats = new HashMap[db.getNumCategoricTarget()];
		for (int i = 0; i < stats.length; i++) {
			stats[i] = new HashMap<String, Integer>();
			
			for (Integer p : pids) {
				if (stats[i].containsKey(db.getPoints().get(p).getCategoricTarget()[i])) {
					int a = stats[i].get(db.getPoints().get(p).getCategoricTarget()[i]) + 1;
					stats[i].put(db.getPoints().get(p).getCategoricTarget()[i], a);
				} else {
					stats[i].put(db.getPoints().get(p).getCategoricTarget()[i], 1);
				}
			}
		}
		
		Map<String, Integer>[] pops = db.getIndividualCategoricTargetCount();
		double sum = 0;
		for (int i = 0; i < stats.length; i++) {
			for (String key : stats[i].keySet()) {
				double a = stats[i].get(key) * 1.0 / pids.size();
				double b = pops[i].get(key) * 1.0 / db.getPoints().size();
				sum += a * Math.log(a / b) / Math.log(2); 
			}
		}
		
		return sum;
	}
	
	public static double totalScore(Database db, List<Integer> pids, int measure, int option) {
		double t = 0;
		if (measure == Constants.KL) {
			t = score(db, pids);
		} else {
			t = scoreIndependence(db, pids);
		}
		
		if (option == Constants.FREQUENCY) { // frequency
			t = pids.size() * t / db.getPoints().size();
		} else { // entropy of frequency
			if (pids.size() < db.getPoints().size()) {
				t = t * Math.sqrt((- (pids.size() * 1.0 / db.getPoints().size()) * Math.log(pids.size() * 1.0 / db.getPoints().size()) / Math.log(2) 
						- ((db.getPoints().size() - pids.size()) * 1.0 / db.getPoints().size()) * Math.log((db.getPoints().size() - pids.size()) * 1.0 / db.getPoints().size()) / Math.log(2)));
			} else {
				t = 0;
			}
		}
		
		return t;
	}
}