package sd.discretizer;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Random;

import sd.computer.HellingerDiv;
import sd.computer.KLDiv;
import sd.computer.MeanStd;
import sd.computer.QuadraticMeasure;
import sd.computer.WRAcc;
import sd.constants.Constants;
import sd.data.*;
import sd.discovery.*;
import sd.score.Score;

public class Discretization {
	private static final int MIN_SIZE = 10;
	private static final int MAX_SIZE = 10000;
	
	public static ArrayList<Subgroup> generateRefinements(Database db, ArrayList<Integer> usedPids, int dim, double e, int method,
														  int measure, int mode, int option, boolean justStart) throws Exception {
		ArrayList<Subgroup> refs = new ArrayList<Subgroup>();
		ArrayList<Bin> bins = null;
		
		int m = db.getPoints().size();
		int mpos = 0;
		double mean = 0;
		double std = 0;
		if (measure == Constants.WRACC || measure == Constants.MEANSTD) {
			ArrayList<Integer> ids = new ArrayList<Integer>();
			for (int i = 0; i < m; i++) {
				ids.add(i);
			}
			
			if (measure == Constants.WRACC) {
				mpos = countPos(db, ids);
			} else {
				mean = MeanStd.mean(db, ids);
				std = MeanStd.std(db, ids, mean);
			}
		}
		
		if (dim < db.getNumNumeric() + db.getNumOrdinal()) {
			int numDesiredBins = Constants.BETA;
			if (method == Constants.SUM) { // sum
				if (measure != Constants.QR || justStart || usedPids.size() < MAX_SIZE) { // if just start or the data size is small
					bins = discretize(db, usedPids, dim, e, method, measure, mode, option, numDesiredBins);
				} else {
					numDesiredBins = 10;
					bins = formInitialBins(db, usedPids, dim, numDesiredBins, method);
					for (Bin b: bins) {
						b.setQuality(db, e, m, mpos, mean, std, measure, mode, option);
					}
				}
			} else { // average
				if (measure != Constants.QR || justStart || usedPids.size() < MAX_SIZE) { // if just start or the data size is small
					bins = discretize2(db, usedPids, dim, e, method, measure, mode, option, numDesiredBins);
				} else {
					numDesiredBins = 10;
					bins = formInitialBins(db, usedPids, dim, numDesiredBins, method);
					for (Bin b: bins) {
						b.setQuality(db, e, m, mpos, mean, std, measure, mode, option);
					}
				}
			}
			
			System.out.printf("number of final bins = %d\n\n", bins.size());
			for (Bin b : bins) {
				Subgroup sub = new Subgroup();
				sub.setPoints(b.getPoints());
				Double min = Double.POSITIVE_INFINITY;
				Double max = Double.NEGATIVE_INFINITY;
				double val;
				for (Integer p : sub.getPoints()) {
					if (dim < db.getNumNumeric()) {
						val = db.getPoints().get(p).getNumeric()[dim];
					} else {
						val = db.getPoints().get(p).getOrdinal()[dim - db.getNumNumeric()];
					}
					
					if (val < min) {
						min = val;
					}
					
					if (val > max) {
						max = val;
					}
				}
				
				Attribute att = new Attribute(dim, min, max);
				sub.getCons().add(att);
				sub.setQuality(b.getQuality());
				refs.add(sub);
			}
		} else {
			HashMap<String, ArrayList<Integer>> map = new HashMap<String, ArrayList<Integer>>();
			String val;
			ArrayList<Integer> pids = null;
			for (Integer p : usedPids) {
				val = db.getPoints().get(p).getCategoric()[dim - db.getNumNumeric() - db.getNumOrdinal()];
				if (map.containsKey(val)) {
					pids = map.get(val);
				} else {
					pids = new ArrayList<Integer>();
				}
				
				pids.add(p);
				map.put(val, pids);
			}
			
			for (String key : map.keySet()) {
				Subgroup sub = new Subgroup();
				sub.setPoints(map.get(key));
				Attribute att = new Attribute(dim, key, key);
				sub.getCons().add(att);
				sub.setQuality(db, e, m, mpos, mean, std, method, measure, mode, option);
				refs.add(sub);
			}
		}
		
		return refs;
	}
	
	@SuppressWarnings("unchecked")
	private static ArrayList<Bin> discretize(Database db, ArrayList<Integer> usedPids, int dim, double e, int method, int measure,
											 int mode, int option, int numDesiredBins) {
		if (Constants.SAMPLING) {
			if (Constants.SUBSAMPLE == null) {
				Constants.SUBSAMPLE = new ArrayList<Integer>();
			} else {
				Constants.SUBSAMPLE.clear();
			}
		}
		ArrayList<Bin> c = formInitialBins(db, usedPids, dim, numDesiredBins, method);
		int beta = c.size();
		
		double[][] phi = null;
		if (measure == Constants.QR) {
			double e1 = e;
			if (Constants.SAMPLING) {
				e1 = QuadraticMeasure.score(db, Constants.SUBSAMPLE, Constants.SUBSAMPLE);
			}
			phi = computePhi(db, c, e1, mode, option);
		} else if (measure == Constants.KL || measure == Constants.KLIND || measure == Constants.HELL) {
			phi = computePhi(db, c, measure, option);
		} else if (measure == Constants.WRACC) {
			phi = computePhiWRAcc(db, c, option);
		} else {
			phi = computePhiMeanStd(db, c);
		}
		for (int i = 0; i < beta; i++) {
			c.get(i).setQuality(phi[i][i]);
		}
		
		double[] qual = new double[beta];
		ArrayList<Bin>[] b = new ArrayList[beta];
		qual[0] = phi[0][0];
		b[0] = new ArrayList<Bin>();
		b[0].add(c.get(0));
		
		for (int i = 1; i < beta; i++) {
			int pos = -1;
			qual[i] = Double.NEGATIVE_INFINITY;
			for (int j = 0; j < i; j++) {
				if (qual[j] + phi[j + 1][i] > qual[i]) {
					qual[i] = qual[j] + phi[j + 1][i];
					pos = j;
				}
			}
			
			b[i] = new ArrayList<Bin>();
			if (pos != -1) {
				b[i].addAll(b[pos]);
			}
			ArrayList<Integer> pids = new ArrayList<Integer>();
			for (int k = pos + 1; k <= i; k++) {
				pids.addAll(c.get(k).getPoints());
			}
			Bin tmp = new Bin(pids);
			tmp.setQuality(phi[pos + 1][i]);
			b[i].add(tmp);
		}
		
		return b[beta - 1];
	}
	
	@SuppressWarnings("unchecked")
	private static ArrayList<Bin> discretize2(Database db, ArrayList<Integer> usedPids, int dim, double e, int method, int measure,
											  int mode, int option, int numDesiredBins) {
		if (Constants.SAMPLING) {
			if (Constants.SUBSAMPLE == null) {
				Constants.SUBSAMPLE = new ArrayList<Integer>();
			} else {
				Constants.SUBSAMPLE.clear();
			}
		}
		ArrayList<Bin> c = formInitialBins(db, usedPids, dim, numDesiredBins, method);
		int beta = c.size();
		
		double[][] phi = null;
		if (measure == Constants.QR) {
			double e1 = e;
			if (Constants.SAMPLING) {
				e1 = QuadraticMeasure.score(db, Constants.SUBSAMPLE, Constants.SUBSAMPLE);
			}
			phi = computePhi(db, c, e1, mode, option);
		} else if (measure == Constants.KL || measure == Constants.KLIND || measure == Constants.HELL) {
			phi = computePhi(db, c, measure, option);
		} else if (measure == Constants.WRACC) {
			phi = computePhiWRAcc(db, c, option);
		} else {
			phi = computePhiMeanStd(db, c);
		}
		for (int i = 0; i < beta; i++) {
			c.get(i).setQuality(phi[i][i]);
		}
		
		double[][] qual = new double[beta + 1][beta];
		ArrayList<Bin>[][] b = new ArrayList[beta + 1][beta];
		Bin tmp = null;
		for (int i = 0; i < beta; i++) {
			b[1][i] = new ArrayList<Bin>();
			if (i == 0) {
				tmp = new Bin(c.get(0).getPoints());
				tmp.setQuality(phi[0][0]);
				b[1][i].add(tmp);
			} else {
				ArrayList<Integer> pids = new ArrayList<Integer>(b[1][i - 1].get(0).getPoints());
				pids.addAll(c.get(i).getPoints());
				tmp = new Bin(pids);
				tmp.setQuality(phi[0][i]);
				b[1][i].add(tmp);
			}
			
			qual[1][i] = phi[0][i];
		}
		
		for (int i = 1; i < beta; i++) {
			int pos = -1;
			double max = Double.NEGATIVE_INFINITY;
			for (int j = 0; j < i; j++) {
				if (qual[1][j] + phi[j + 1][i] > max) {
					max = qual[1][j] + phi[j + 1][i];
					pos = j;
				}
			}
			
			qual[2][i] = qual[1][pos] + phi[pos + 1][i];
			
			b[2][i] = new ArrayList<Bin>();
			b[2][i].addAll(b[1][pos]);
			ArrayList<Integer> pids = new ArrayList<Integer>();
			for (int k = pos + 1; k <= i; k++) {
				pids.addAll(c.get(k).getPoints());
			}
			tmp = new Bin(pids);
			tmp.setQuality(phi[pos + 1][i]);
			b[2][i].add(tmp);
		}
		
		for (int delta = 3; delta <= beta; delta++) {
			for (int i = delta - 1; i < beta; i++) {
				int pos = -1;
				double max = Double.NEGATIVE_INFINITY;
				for (int j = delta - 2; j < i; j++) {
					if (qual[delta - 1][j] + phi[j + 1][i] > max) {
						max = qual[delta - 1][j] + phi[j + 1][i];
						pos = j;
					}
				}
				
				qual[delta][i] = qual[delta - 1][pos] + phi[pos + 1][i];
				
				b[delta][i] = new ArrayList<Bin>();
				//System.out.printf("pos = %d\n", pos);
				b[delta][i].addAll(b[delta - 1][pos]);
				ArrayList<Integer> pids = new ArrayList<Integer>();
				for (int k = pos + 1; k <= i; k++) {
					pids.addAll(c.get(k).getPoints());
				}
				tmp = new Bin(pids);
				tmp.setQuality(phi[pos + 1][i]);
				b[delta][i].add(tmp);
			}
		}
		
		int deltaMax = -1;
		double max = Double.NEGATIVE_INFINITY;
		int start = 2;
		if (beta == 1) {
			start = 1;
		}
		for (int delta = start; delta <= beta; delta++) {
			if (qual[delta][beta - 1] / delta > max) {
				max = qual[delta][beta - 1] / delta;
				deltaMax = delta;
			}
		}
		
		return b[deltaMax][beta - 1];
	}
	
	private static ArrayList<Bin> formInitialBins(Database db, ArrayList<Integer> pids, int dim, int numDesiredBins, int method) {
		ArrayList<SortedKey> sortedKeys = new ArrayList<SortedKey>();
		SortedKey key = null;
		for (Integer p : pids) {
			key = new SortedKey();
			if (dim < db.getNumNumeric()) {
				key.value = db.getPoints().get(p).getNumeric()[dim];
			} else {
				key.value = db.getPoints().get(p).getOrdinal()[dim - db.getNumNumeric()];
			}
			key.pid = p;
			sortedKeys.add(key);
		}
		//Collections.sort(sortedKeys);
		return formInitialBins(sortedKeys, numDesiredBins);
	}
	
	private static ArrayList<Bin> formInitialBins(ArrayList<SortedKey> keys, int numDesiredBins) {
		ArrayList<Bin> ret = new ArrayList<Bin>();
		//System.out.printf("numDesiredBins = %d\n", numDesiredBins);
		
		// extract distinct values
		HashMap<Double, ArrayList<Integer>> distinctValues = new HashMap<Double, ArrayList<Integer>>();
		ArrayList<Double> sortedKeys = new ArrayList<Double>();
		for (SortedKey key : keys) {
			ArrayList<Integer> pids = null;
			if (distinctValues.containsKey(key.value)) {
				pids = distinctValues.get(key.value);
			} else {
				sortedKeys.add(key.value);
				pids = new ArrayList<Integer>();
			}
			
			pids.add(key.pid);
			distinctValues.put(key.value, pids);
		}
		Collections.sort(sortedKeys);
		
		// number of distinct values per bin
		int numRows = keys.size();
		int totalNumBins = numDesiredBins;
		int desiredBinSize = (int)Math.floor(numRows * 1.0 / totalNumBins);
		//System.out.printf("numDesiredBins = %d\n", numDesiredBins);
		
		// discretize
		int[] indices = new int[distinctValues.size()];
		int i = 1;
		int curDistinctVal = 0;
		int curBinIndex = 1;
		int count;
		while (curDistinctVal < indices.length) {
			count = 0;
			for (int j = 0; j < indices.length; j++) {
				if (indices[j] == curBinIndex) {
					count = count + distinctValues.get(sortedKeys.get(j)).size();
				}
			}
			
			if (count != 0 && Math.abs(count + distinctValues.get(sortedKeys.get(curDistinctVal)).size() - desiredBinSize) >= Math.abs(count - desiredBinSize)) {
				curBinIndex++;
				desiredBinSize = (int)((numRows - i + 1) * 1.0) / (totalNumBins - curBinIndex + 1);
			}
			indices[curDistinctVal] = curBinIndex;
			i = i + distinctValues.get(sortedKeys.get(curDistinctVal)).size();
			curDistinctVal++;
		}
		//System.out.println("curBinIndex = " + curBinIndex);
		
		// create bin
		Bin tmpBin = null;
		curBinIndex = 1;
		curDistinctVal = 0;
		int newDistinctVal;
		ArrayList<Integer> ds = null;
		while (curDistinctVal < indices.length) {
			for (i = curDistinctVal + 1; i < indices.length; i++) {
				if (indices[i] != curBinIndex) {
					break;
				}
			}
			newDistinctVal = i;
			
			// get current value and add the corresponding points to the bin
			ds = new ArrayList<Integer>();
			for (i = curDistinctVal; i < newDistinctVal; i++) {
				ds.addAll(distinctValues.get(sortedKeys.get(i)));
			}
			tmpBin = new Bin(ds);
			
			// add the bin to the list of bins
			ret.add(tmpBin);
			
			curDistinctVal = newDistinctVal;
			curBinIndex++;
		} // end for
		
		// subsampling and control check
		count = 0;
		for (Bin bin : ret) {
			if (Constants.SAMPLING) {
				bin.setSubPoints(subsample(bin.getPoints(), Constants.EPSILON));
				if (Constants.SUBSAMPLE != null) {
					Constants.SUBSAMPLE.addAll(bin.getSubPoints());
				}
			}
			
			count += bin.getPoints().size();
		}
		if (count != numRows) {
			throw new ArithmeticException("some point is missing!");
		}
		
		System.out.printf("number of initial bins = %d\n", ret.size());
			
		return ret;
	}
	
	private static double[][] computePhi(Database db, ArrayList<Bin> c, double e, int mode, int option) {
		int beta = c.size();
		double[][] phi = new double[beta][beta];
		Score[][] phiComps = new Score[beta][beta];
		double[][] inter = new double[beta][beta];
		double[][] interSum = new double[beta][beta];
		
		for (int i = 0; i < beta; i++) {
			if (Constants.SAMPLING) {
				phiComps[i][i] = Score.score(db, new Bin(c.get(i).getSubPoints()));
			} else {
				phiComps[i][i] = Score.score(db, c.get(i));
			}
			phi[i][i] = QuadraticMeasure.totalScore(phiComps[i][i], e, mode, option);
		
			for (int j = i + 1; j < beta; j++) {
				if (Constants.SAMPLING) {
					inter[i][j] = QuadraticMeasure.score(db, c.get(i).getSubPoints(), c.get(j).getSubPoints());
				} else {
					inter[i][j] = QuadraticMeasure.score(db, c.get(i).getPoints(), c.get(j).getPoints());
				}
			}
		}
		
		for (int i = 1; i < beta; i++) {
			double sum = 0;
			for (int j = i - 1; j >= 0; j--) {
				sum += inter[j][i];
				interSum[j][i] = sum; 
			}
		}
		
		for (int i = 0; i < beta; i++) {
			for (int j = 0; j < i; j++) {
				//if (j == 0 && i == beta - 1) {
				//	double a = phi[j][i];
				//	System.out.printf("a = %f\n", a);
				//}
				
				phiComps[j][i] = new Score();
				phiComps[j][i].first = phiComps[j][i - 1].first + phiComps[i][i].first + 2 * interSum[j][i];
				//double x = phiComps[j][i].first;
				phiComps[j][i].second = phiComps[j][i - 1].second + phiComps[i][i].second;
				//double y = phiComps[j][i].second;
				phiComps[j][i].s = phiComps[j][i - 1].s + phiComps[i][i].s;
				phiComps[j][i].m = phiComps[i][i].m;
				phi[j][i] = QuadraticMeasure.totalScore(phiComps[j][i], e, mode, option);
				//double z = phi[j][i];
				
				if (phi[j][i] < 0) {
					//throw new ArithmeticException(String.format("phi[j][i] = %f", phi[j][i]));
					phi[j][i] = 0;
				}
			}
		}
		
		return phi;
	}
	
	private static double[][] computePhi(Database db, ArrayList<Bin> c, int measure, int option) {
		int beta = c.size();
		double[][] phi = new double[beta][beta];
		
		for (int i = 0; i < beta; i++) {
			if (measure != Constants.HELL) {
				phi[i][i] = KLDiv.totalScore(db, c.get(i).getPoints(), measure, option);
			} else {
				phi[i][i] = HellingerDiv.totalScore(db, c.get(i).getPoints(), option);
			}
		}
		
		for (int i = 0; i < beta; i++) {
			ArrayList<Integer> pids = new ArrayList<Integer>();
			pids.addAll(c.get(i).getPoints());
			for (int j = i - 1; j >= 0; j--) {
				pids.addAll(c.get(j).getPoints());
				if (measure != Constants.HELL) {
					phi[j][i] = KLDiv.totalScore(db, pids, measure, option);
				} else {
					phi[j][i] = HellingerDiv.totalScore(db, pids, option);
				}
			}
		}
		
		return phi;
	}
	
	private static double[][] computePhiWRAcc(Database db, ArrayList<Bin> c, int option) {
		int m = db.getPoints().size();
		ArrayList<Integer> ids = new ArrayList<Integer>();
		for (int i = 0; i < m; i++) {
			ids.add(i);
		}
		int mpos = countPos(db, ids);
		
		int beta = c.size();
		double[][] phi = new double[beta][beta];
		int[] binpos = new int[beta];
		for (int i = 0; i < beta; i++) {
			binpos[i] = countPos(db, c.get(i).getPoints());
		}
		
		for (int i = 0; i < beta; i++) {
			phi[i][i] = WRAcc.score(c.get(i).getPoints().size(), binpos[i], m, mpos, option);
		}
		
		
		for (int i = 0; i < beta; i++) {
			int s = c.get(i).getPoints().size();
			int spos = binpos[i];
			for (int j = i - 1; j >= 0; j--) {
				s += c.get(j).getPoints().size();
				spos += binpos[j];
				phi[j][i] = WRAcc.score(s, spos, m, mpos, option);
			}
		}
		
		return phi;
	}
	
	private static double[][] computePhiMeanStd(Database db, ArrayList<Bin> c) {
		int m = db.getPoints().size();
		ArrayList<Integer> ids = new ArrayList<Integer>();
		for (int i = 0; i < m; i++) {
			ids.add(i);
		}
		double mean = MeanStd.mean(db, ids);
		double std = MeanStd.std(db, ids, mean);
		
		int beta = c.size();
		double[][] phi = new double[beta][beta];
		double[] binmean = new double[beta];
		for (int i = 0; i < beta; i++) {
			binmean[i] = MeanStd.mean(db, c.get(i).getPoints());
		}
		
		for (int i = 0; i < beta; i++) {
			phi[i][i] = MeanStd.score(c.get(i).getPoints().size(), binmean[i], mean, std);
		}
		
		
		for (int i = 0; i < beta; i++) {
			int s = c.get(i).getPoints().size();
			double smean = binmean[i];
			for (int j = i - 1; j >= 0; j--) {
				smean += s * smean + c.get(j).getPoints().size() * binmean[j];
				s += c.get(j).getPoints().size();
				smean = smean / s;
				phi[j][i] = MeanStd.score(s, smean, mean, std);
			}
		}
		
		return phi;
	}
	
	public static int countPos(Database db, ArrayList<Integer> pids) {
		int count = 0;
		for (Integer p : pids) {
			if (db.getPoints().get(p).getCategoricTarget()[0].equals("1")) {
				count++;
			}
		}
		
		return count;
	}
	
	public static ArrayList<Integer> subsample(ArrayList<Integer> pids, double epsilon) {
		if (pids.size() < MIN_SIZE) {
			return pids;
		}
		
		ArrayList<Integer> ret = new ArrayList<Integer>();
		int sampleSize = (int)(pids.size() * epsilon);
		Random r = new Random();
		for (int i = 0; i < sampleSize; i++) {
			ret.add(pids.get(r.nextInt(pids.size())));
		}
		
		return ret;
	}
}